package java.awt;

import sun.awt.AWTAccessor;
import java.util.ListIterator;
import java.util.Set;
import java.awt.peer.ComponentPeer;
import java.util.Iterator;
import java.util.List;
import java.awt.peer.LightweightPeer;
import sun.awt.TimedWindowEvent;
import sun.awt.AppContext;
import sun.awt.SunToolkit;
import java.awt.event.WindowEvent;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.util.LinkedList;
import java.lang.ref.WeakReference;
import sun.util.logging.PlatformLogger;

public class DefaultKeyboardFocusManager extends KeyboardFocusManager
{
    private static final PlatformLogger focusLog;
    private static final WeakReference<Window> NULL_WINDOW_WR;
    private static final WeakReference<Component> NULL_COMPONENT_WR;
    private WeakReference<Window> realOppositeWindowWR;
    private WeakReference<Component> realOppositeComponentWR;
    private int inSendMessage;
    private LinkedList<KeyEvent> enqueuedKeyEvents;
    private LinkedList<DefaultKeyboardFocusManager.TypeAheadMarker> typeAheadMarkers;
    private boolean consumeNextKeyTyped;
    private Component restoreFocusTo;
    
    public DefaultKeyboardFocusManager() {
        this.realOppositeWindowWR = DefaultKeyboardFocusManager.NULL_WINDOW_WR;
        this.realOppositeComponentWR = DefaultKeyboardFocusManager.NULL_COMPONENT_WR;
        this.enqueuedKeyEvents = new LinkedList();
        this.typeAheadMarkers = new LinkedList();
    }
    
    private Window getOwningFrameDialog(Window window) {
        while (window != null && !(window instanceof Frame) && !(window instanceof Dialog)) {
            window = (Window)window.getParent();
        }
        return window;
    }
    
    private void restoreFocus(FocusEvent focusEvent, Window window) {
        Component component = (Component)this.realOppositeComponentWR.get();
        Component component2 = focusEvent.getComponent();
        if ((window == null || !this.restoreFocus(window, component2, false)) && (component == null || !this.doRestoreFocus(component, component2, false)) && (focusEvent.getOppositeComponent() == null || !this.doRestoreFocus(focusEvent.getOppositeComponent(), component2, false))) {
            this.clearGlobalFocusOwnerPriv();
        }
    }
    
    private void restoreFocus(WindowEvent windowEvent) {
        Window window = (Window)this.realOppositeWindowWR.get();
        if ((window == null || !this.restoreFocus(window, null, false)) && (windowEvent.getOppositeWindow() == null || !this.restoreFocus(windowEvent.getOppositeWindow(), null, false))) {
            this.clearGlobalFocusOwnerPriv();
        }
    }
    
    private boolean restoreFocus(Window window, Component component, boolean b) {
        this.restoreFocusTo = null;
        Component restoreFocusTo = KeyboardFocusManager.getMostRecentFocusOwner(window);
        if (restoreFocusTo != null && restoreFocusTo != component) {
            if (KeyboardFocusManager.getHeavyweight(window) != this.getNativeFocusOwner()) {
                if (!restoreFocusTo.isShowing() || !restoreFocusTo.canBeFocusOwner()) {
                    restoreFocusTo = restoreFocusTo.getNextFocusCandidate();
                }
                if (restoreFocusTo != null && restoreFocusTo != component) {
                    if (!restoreFocusTo.requestFocus(false, CausedFocusEvent.Cause.ROLLBACK)) {
                        this.restoreFocusTo = restoreFocusTo;
                    }
                    return true;
                }
            }
            else if (this.doRestoreFocus(restoreFocusTo, component, false)) {
                return true;
            }
        }
        if (b) {
            this.clearGlobalFocusOwnerPriv();
            return true;
        }
        return false;
    }
    
    private boolean restoreFocus(Component component, boolean b) {
        return this.doRestoreFocus(component, null, b);
    }
    
    private boolean doRestoreFocus(Component component, Component component2, boolean b) {
        boolean requestFocus = true;
        if (component != component2 && component.isShowing() && component.canBeFocusOwner() && (requestFocus = component.requestFocus(false, CausedFocusEvent.Cause.ROLLBACK))) {
            return true;
        }
        if (requestFocus == 0 && this.getGlobalFocusedWindow() != SunToolkit.getContainingWindow(component)) {
            this.restoreFocusTo = component;
            return true;
        }
        Component nextFocusCandidate = component.getNextFocusCandidate();
        if (nextFocusCandidate != null && nextFocusCandidate != component2 && nextFocusCandidate.requestFocusInWindow(CausedFocusEvent.Cause.ROLLBACK)) {
            return true;
        }
        if (b) {
            this.clearGlobalFocusOwnerPriv();
            return true;
        }
        return false;
    }
    
    static boolean sendMessage(Component component, AWTEvent awtEvent) {
        awtEvent.isPosted = true;
        AppContext appContext = AppContext.getAppContext();
        final AppContext appContext2 = component.appContext;
        final DefaultKeyboardFocusManager.DefaultKeyboardFocusManagerSentEvent event = new DefaultKeyboardFocusManager.DefaultKeyboardFocusManagerSentEvent(awtEvent, appContext);
        if (appContext == appContext2) {
            event.dispatch();
        }
        else {
            if (appContext2.isDisposed()) {
                return false;
            }
            SunToolkit.postEvent(appContext2, event);
            if (EventQueue.isDispatchThread()) {
                ((EventDispatchThread)Thread.currentThread()).pumpEvents(1007, new Conditional() {
                    @Override
                    public boolean evaluate() {
                        return !event.dispatched && !appContext2.isDisposed();
                    }
                });
            }
            else {
                synchronized (event) {
                    while (!event.dispatched && !appContext2.isDisposed()) {
                        try {
                            event.wait(1000L);
                        }
                        catch (InterruptedException ex) {
                            break;
                        }
                    }
                }
            }
        }
        return event.dispatched;
    }
    
    private boolean repostIfFollowsKeyEvents(WindowEvent windowEvent) {
        if (!(windowEvent instanceof TimedWindowEvent)) {
            return false;
        }
        long when = ((TimedWindowEvent)windowEvent).getWhen();
        synchronized (this) {
            KeyEvent keyEvent = this.enqueuedKeyEvents.isEmpty() ? null : ((KeyEvent)this.enqueuedKeyEvents.getFirst());
            if (keyEvent != null && when >= keyEvent.getWhen()) {
                DefaultKeyboardFocusManager.TypeAheadMarker typeAheadMarker = this.typeAheadMarkers.isEmpty() ? null : ((DefaultKeyboardFocusManager.TypeAheadMarker)this.typeAheadMarkers.getFirst());
                if (typeAheadMarker != null) {
                    Window containingWindow = typeAheadMarker.untilFocused.getContainingWindow();
                    if (containingWindow != null && containingWindow.isFocused()) {
                        SunToolkit.postEvent(AppContext.getAppContext(), new SequencedEvent(windowEvent));
                        return true;
                    }
                }
            }
        }
        return false;
    }
    
    @Override
    public boolean dispatchEvent(AWTEvent obj) {
        if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINE) && (obj instanceof WindowEvent || obj instanceof FocusEvent)) {
            DefaultKeyboardFocusManager.focusLog.fine("" + obj);
        }
        switch (obj.getID()) {
            case 207: {
                if (this.repostIfFollowsKeyEvents((WindowEvent)obj)) {
                    break;
                }
                WindowEvent windowEvent = (AWTEvent)obj;
                Window globalFocusedWindow = this.getGlobalFocusedWindow();
                Window window = windowEvent.getWindow();
                if (window == globalFocusedWindow) {
                    break;
                }
                if (!window.isFocusableWindow() || !window.isVisible() || !window.isDisplayable()) {
                    this.restoreFocus(windowEvent);
                    break;
                }
                if (globalFocusedWindow != null && !sendMessage(globalFocusedWindow, new WindowEvent(globalFocusedWindow, 208, window))) {
                    this.setGlobalFocusOwner(null);
                    this.setGlobalFocusedWindow(null);
                }
                Window owningFrameDialog = this.getOwningFrameDialog(window);
                Window globalActiveWindow = this.getGlobalActiveWindow();
                if (owningFrameDialog != globalActiveWindow) {
                    sendMessage(owningFrameDialog, new WindowEvent(owningFrameDialog, 205, globalActiveWindow));
                    if (owningFrameDialog != this.getGlobalActiveWindow()) {
                        this.restoreFocus(windowEvent);
                        break;
                    }
                }
                this.setGlobalFocusedWindow(window);
                if (window != this.getGlobalFocusedWindow()) {
                    this.restoreFocus(windowEvent);
                    break;
                }
                if (this.inSendMessage == 0) {
                    Component component = KeyboardFocusManager.getMostRecentFocusOwner(window);
                    int n = this.restoreFocusTo != null && component == this.restoreFocusTo;
                    if (component == null && window.isFocusableWindow()) {
                        component = window.getFocusTraversalPolicy().getInitialComponent(window);
                    }
                    Component setTemporaryLostComponent = null;
                    synchronized (KeyboardFocusManager.class) {
                        setTemporaryLostComponent = window.setTemporaryLostComponent(null);
                    }
                    if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
                        DefaultKeyboardFocusManager.focusLog.finer("tempLost {0}, toFocus {1}", new Object[] { setTemporaryLostComponent, component });
                    }
                    if (setTemporaryLostComponent != null) {
                        setTemporaryLostComponent.requestFocusInWindow((n && setTemporaryLostComponent == component) ? CausedFocusEvent.Cause.ROLLBACK : CausedFocusEvent.Cause.ACTIVATION);
                    }
                    if (component != null && component != setTemporaryLostComponent) {
                        component.requestFocusInWindow(CausedFocusEvent.Cause.ACTIVATION);
                    }
                }
                this.restoreFocusTo = null;
                Window opposite = (Window)this.realOppositeWindowWR.get();
                if (opposite != windowEvent.getOppositeWindow()) {
                    windowEvent = new WindowEvent(window, 207, opposite);
                }
                return this.typeAheadAssertions(window, windowEvent);
            }
            case 205: {
                WindowEvent windowEvent2 = (WindowEvent)obj;
                Window globalActiveWindow2 = this.getGlobalActiveWindow();
                Window window2 = windowEvent2.getWindow();
                if (globalActiveWindow2 == window2) {
                    break;
                }
                if (globalActiveWindow2 != null) {
                    if (!sendMessage(globalActiveWindow2, new WindowEvent(globalActiveWindow2, 206, window2))) {
                        this.setGlobalActiveWindow(null);
                    }
                    if (this.getGlobalActiveWindow() != null) {
                        break;
                    }
                }
                this.setGlobalActiveWindow(window2);
                if (window2 != this.getGlobalActiveWindow()) {
                    break;
                }
                return this.typeAheadAssertions(window2, windowEvent2);
            }
            case 1004: {
                this.restoreFocusTo = null;
                FocusEvent focusEvent = (FocusEvent)obj;
                CausedFocusEvent.Cause cause = (focusEvent instanceof CausedFocusEvent) ? ((CausedFocusEvent)focusEvent).getCause() : CausedFocusEvent.Cause.UNKNOWN;
                Component globalFocusOwner = this.getGlobalFocusOwner();
                Component component2 = focusEvent.getComponent();
                if (globalFocusOwner == component2) {
                    if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINE)) {
                        DefaultKeyboardFocusManager.focusLog.fine("Skipping {0} because focus owner is the same", new Object[] { obj });
                    }
                    this.dequeueKeyEvents(-1L, component2);
                    break;
                }
                else {
                    if (globalFocusOwner != null && !sendMessage(globalFocusOwner, new CausedFocusEvent(globalFocusOwner, 1005, focusEvent.isTemporary(), component2, cause))) {
                        this.setGlobalFocusOwner(null);
                        if (!focusEvent.isTemporary()) {
                            this.setGlobalPermanentFocusOwner(null);
                        }
                    }
                    Window containingWindow = SunToolkit.getContainingWindow(component2);
                    Window globalFocusedWindow2 = this.getGlobalFocusedWindow();
                    if (containingWindow != null && containingWindow != globalFocusedWindow2) {
                        sendMessage(containingWindow, new WindowEvent(containingWindow, 207, globalFocusedWindow2));
                        if (containingWindow != this.getGlobalFocusedWindow()) {
                            this.dequeueKeyEvents(-1L, component2);
                            break;
                        }
                    }
                    if (!component2.isFocusable() || !component2.isShowing() || (!component2.isEnabled() && !cause.equals(CausedFocusEvent.Cause.UNKNOWN))) {
                        this.dequeueKeyEvents(-1L, component2);
                        if (KeyboardFocusManager.isAutoFocusTransferEnabled()) {
                            if (containingWindow == null) {
                                this.restoreFocus(focusEvent, globalFocusedWindow2);
                            }
                            else {
                                this.restoreFocus(focusEvent, containingWindow);
                            }
                            KeyboardFocusManager.setMostRecentFocusOwner(containingWindow, null);
                            break;
                        }
                        else {
                            break;
                        }
                    }
                    else {
                        this.setGlobalFocusOwner(component2);
                        if (component2 != this.getGlobalFocusOwner()) {
                            this.dequeueKeyEvents(-1L, component2);
                            if (KeyboardFocusManager.isAutoFocusTransferEnabled()) {
                                this.restoreFocus(focusEvent, containingWindow);
                                break;
                            }
                            break;
                        }
                        else {
                            if (!focusEvent.isTemporary()) {
                                this.setGlobalPermanentFocusOwner(component2);
                                if (component2 != this.getGlobalPermanentFocusOwner()) {
                                    this.dequeueKeyEvents(-1L, component2);
                                    if (KeyboardFocusManager.isAutoFocusTransferEnabled()) {
                                        this.restoreFocus(focusEvent, containingWindow);
                                        break;
                                    }
                                    break;
                                }
                            }
                            this.setNativeFocusOwner(KeyboardFocusManager.getHeavyweight(component2));
                            Component opposite2 = (Component)this.realOppositeComponentWR.get();
                            if (opposite2 != null && opposite2 != focusEvent.getOppositeComponent()) {
                                focusEvent = new CausedFocusEvent(component2, 1004, focusEvent.isTemporary(), opposite2, cause);
                                focusEvent.isPosted = true;
                            }
                            return this.typeAheadAssertions(component2, focusEvent);
                        }
                    }
                }
            }
            case 1005: {
                FocusEvent focusEvent2 = (FocusEvent)obj;
                Component globalFocusOwner2 = this.getGlobalFocusOwner();
                if (globalFocusOwner2 == null) {
                    if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINE)) {
                        DefaultKeyboardFocusManager.focusLog.fine("Skipping {0} because focus owner is null", new Object[] { obj });
                        break;
                    }
                    break;
                }
                else if (globalFocusOwner2 == focusEvent2.getOppositeComponent()) {
                    if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINE)) {
                        DefaultKeyboardFocusManager.focusLog.fine("Skipping {0} because current focus owner is equal to opposite", new Object[] { obj });
                        break;
                    }
                    break;
                }
                else {
                    this.setGlobalFocusOwner(null);
                    if (this.getGlobalFocusOwner() != null) {
                        this.restoreFocus(globalFocusOwner2, true);
                        break;
                    }
                    if (!focusEvent2.isTemporary()) {
                        this.setGlobalPermanentFocusOwner(null);
                        if (this.getGlobalPermanentFocusOwner() != null) {
                            this.restoreFocus(globalFocusOwner2, true);
                            break;
                        }
                    }
                    else {
                        Window containingWindow2 = globalFocusOwner2.getContainingWindow();
                        if (containingWindow2 != null) {
                            containingWindow2.setTemporaryLostComponent(globalFocusOwner2);
                        }
                    }
                    this.setNativeFocusOwner(null);
                    focusEvent2.setSource(globalFocusOwner2);
                    this.realOppositeComponentWR = ((focusEvent2.getOppositeComponent() != null) ? new WeakReference(globalFocusOwner2) : DefaultKeyboardFocusManager.NULL_COMPONENT_WR);
                    return this.typeAheadAssertions(globalFocusOwner2, focusEvent2);
                }
            }
            case 206: {
                WindowEvent windowEvent3 = (WindowEvent)obj;
                Window globalActiveWindow3 = this.getGlobalActiveWindow();
                if (globalActiveWindow3 == null) {
                    break;
                }
                if (globalActiveWindow3 != obj.getSource()) {
                    break;
                }
                this.setGlobalActiveWindow(null);
                if (this.getGlobalActiveWindow() != null) {
                    break;
                }
                windowEvent3.setSource(globalActiveWindow3);
                return this.typeAheadAssertions(globalActiveWindow3, windowEvent3);
            }
            case 208: {
                if (this.repostIfFollowsKeyEvents((WindowEvent)obj)) {
                    break;
                }
                WindowEvent windowEvent4 = (WindowEvent)obj;
                Window globalFocusedWindow3 = this.getGlobalFocusedWindow();
                Window window3 = windowEvent4.getWindow();
                Window globalActiveWindow4 = this.getGlobalActiveWindow();
                Window opposite3 = windowEvent4.getOppositeWindow();
                if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINE)) {
                    DefaultKeyboardFocusManager.focusLog.fine("Active {0}, Current focused {1}, losing focus {2} opposite {3}", new Object[] { globalActiveWindow4, globalFocusedWindow3, window3, opposite3 });
                }
                if (globalFocusedWindow3 == null) {
                    break;
                }
                if (this.inSendMessage == 0 && window3 == globalActiveWindow4 && opposite3 == globalFocusedWindow3) {
                    break;
                }
                Component globalFocusOwner3 = this.getGlobalFocusOwner();
                if (globalFocusOwner3 != null) {
                    opposite3 = null;
                    if (opposite3 != null) {
                        opposite3 = opposite3.getTemporaryLostComponent();
                        if (opposite3 == null) {
                            opposite3 = opposite3.getMostRecentFocusOwner();
                        }
                    }
                    if (opposite3 == null) {
                    }
                    sendMessage(globalFocusOwner3, new CausedFocusEvent(globalFocusOwner3, 1005, true, opposite3, CausedFocusEvent.Cause.ACTIVATION));
                }
                this.setGlobalFocusedWindow(null);
                if (this.getGlobalFocusedWindow() != null) {
                    this.restoreFocus(globalFocusedWindow3, null, true);
                    break;
                }
                windowEvent4.setSource(globalFocusedWindow3);
                this.realOppositeWindowWR = ((opposite3 != null) ? new WeakReference(globalFocusedWindow3) : DefaultKeyboardFocusManager.NULL_WINDOW_WR);
                this.typeAheadAssertions(globalFocusedWindow3, windowEvent4);
                if (opposite3 == null && globalActiveWindow4 != null) {
                    sendMessage(globalActiveWindow4, new WindowEvent(globalActiveWindow4, 206, null));
                    if (this.getGlobalActiveWindow() != null) {
                        this.restoreFocus(globalFocusedWindow3, null, true);
                        break;
                    }
                    break;
                }
                else {
                    break;
                }
            }
            case 400:
            case 401:
            case 402: {
                return this.typeAheadAssertions(null, obj);
            }
            default: {
                return false;
            }
        }
        return true;
    }
    
    @Override
    public boolean dispatchKeyEvent(KeyEvent keyEvent) {
        Component component = keyEvent.isPosted ? this.getFocusOwner() : keyEvent.getComponent();
        if (component != null && component.isShowing() && component.canBeFocusOwner() && !keyEvent.isConsumed()) {
            Component component2 = keyEvent.getComponent();
            if (component2 != null && component2.isEnabled()) {
                this.redispatchEvent(component2, keyEvent);
            }
        }
        boolean postProcessKeyEvent = false;
        List<KeyEventPostProcessor> keyEventPostProcessors = this.getKeyEventPostProcessors();
        if (keyEventPostProcessors != null) {
            for (Iterator iterator = keyEventPostProcessors.iterator(); postProcessKeyEvent == 0 && iterator.hasNext(); postProcessKeyEvent = ((KeyEventPostProcessor)iterator.next()).postProcessKeyEvent(keyEvent)) {}
        }
        if (postProcessKeyEvent == 0) {
            this.postProcessKeyEvent(keyEvent);
        }
        Component component3 = keyEvent.getComponent();
        ComponentPeer componentPeer = component3.getPeer();
        if (componentPeer == null || componentPeer instanceof LightweightPeer) {
            Container nativeContainer = component3.getNativeContainer();
            if (nativeContainer != null) {
                componentPeer = nativeContainer.getPeer();
            }
        }
        if (componentPeer != null) {
            componentPeer.handleEvent(keyEvent);
        }
        return true;
    }
    
    @Override
    public boolean postProcessKeyEvent(KeyEvent keyEvent) {
        if (!keyEvent.isConsumed()) {
            Component component = keyEvent.getComponent();
            Container container = (Container)(Container)((component instanceof Container) ? component : component.getParent());
            if (container != null) {
                container.postProcessKeyEvent(keyEvent);
            }
        }
        return true;
    }
    
    private void pumpApprovedKeyEvents() {
        KeyEvent keyEvent;
        do {
            keyEvent = null;
            synchronized (this) {
                if (this.enqueuedKeyEvents.size() != 0) {
                    keyEvent = (KeyEvent)this.enqueuedKeyEvents.getFirst();
                    if (this.typeAheadMarkers.size() != 0 && keyEvent.getWhen() > ((DefaultKeyboardFocusManager.TypeAheadMarker)this.typeAheadMarkers.getFirst()).after) {
                        keyEvent = null;
                    }
                    if (keyEvent != null) {
                        if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
                            DefaultKeyboardFocusManager.focusLog.finer("Pumping approved event {0}", new Object[] { keyEvent });
                        }
                        this.enqueuedKeyEvents.removeFirst();
                    }
                }
            }
            if (keyEvent != null) {
                this.preDispatchKeyEvent(keyEvent);
            }
            else {
                continue;
            }
        } while (keyEvent != null);
    }
    
    void dumpMarkers() {
        if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
            DefaultKeyboardFocusManager.focusLog.finest(">>> Markers dump, time: {0}", new Object[] { Long.valueOf(System.currentTimeMillis()) });
            synchronized (this) {
                if (this.typeAheadMarkers.size() != 0) {
                    Iterator<Object> iterator = this.typeAheadMarkers.iterator();
                    while (iterator.hasNext()) {
                        DefaultKeyboardFocusManager.focusLog.finest("    {0}", new Object[] { (Object)iterator.next() });
                    }
                }
            }
        }
    }
    
    private boolean typeAheadAssertions(Component component, AWTEvent awtEvent) {
        this.pumpApprovedKeyEvents();
        switch (awtEvent.getID()) {
            case 400:
            case 401:
            case 402: {
                KeyEvent e = (KeyEvent)awtEvent;
                synchronized (this) {
                    if (awtEvent.isPosted && this.typeAheadMarkers.size() != 0) {
                        DefaultKeyboardFocusManager.TypeAheadMarker typeAheadMarker = (DefaultKeyboardFocusManager.TypeAheadMarker)this.typeAheadMarkers.getFirst();
                        if (e.getWhen() > typeAheadMarker.after) {
                            if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
                                DefaultKeyboardFocusManager.focusLog.finer("Storing event {0} because of marker {1}", new Object[] { e, typeAheadMarker });
                            }
                            this.enqueuedKeyEvents.addLast(e);
                            return true;
                        }
                    }
                }
                return this.preDispatchKeyEvent(e);
            }
            case 1004: {
                if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINEST)) {
                    DefaultKeyboardFocusManager.focusLog.finest("Markers before FOCUS_GAINED on {0}", new Object[] { component });
                }
                this.dumpMarkers();
                synchronized (this) {
                    boolean n = false;
                    if (this.hasMarker(component)) {
                        Iterator<Object> iterator = this.typeAheadMarkers.iterator();
                        while (iterator.hasNext()) {
                            if (((DefaultKeyboardFocusManager.TypeAheadMarker)iterator.next()).untilFocused == component) {
                                n = true;
                            }
                            else if (!(n == 0)) {
                                break;
                            }
                            iterator.remove();
                        }
                    }
                    else if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
                        DefaultKeyboardFocusManager.focusLog.finer("Event without marker {0}", new Object[] { awtEvent });
                    }
                }
                DefaultKeyboardFocusManager.focusLog.finest("Markers after FOCUS_GAINED");
                this.dumpMarkers();
                this.redispatchEvent(component, awtEvent);
                this.pumpApprovedKeyEvents();
                return true;
            }
            default: {
                this.redispatchEvent(component, awtEvent);
                return true;
            }
        }
    }
    
    private boolean hasMarker(Component component) {
        Iterator<Object> iterator = this.typeAheadMarkers.iterator();
        while (iterator.hasNext()) {
            if (((DefaultKeyboardFocusManager.TypeAheadMarker)iterator.next()).untilFocused != component) {
                continue;
            }
            return true;
        }
        return false;
    }
    
    @Override
    void clearMarkers() {
        synchronized (this) {
            this.typeAheadMarkers.clear();
        }
    }
    
    private boolean preDispatchKeyEvent(KeyEvent currentEventAndMostRecentTime) {
        if (currentEventAndMostRecentTime.isPosted) {
            Component focusOwner = this.getFocusOwner();
            currentEventAndMostRecentTime.setSource((focusOwner != null) ? focusOwner : this.getFocusedWindow());
        }
        if (currentEventAndMostRecentTime.getSource() == null) {
            return true;
        }
        EventQueue.setCurrentEventAndMostRecentTime(currentEventAndMostRecentTime);
        if (KeyboardFocusManager.isProxyActive(currentEventAndMostRecentTime)) {
            Container nativeContainer = ((Component)currentEventAndMostRecentTime.getSource()).getNativeContainer();
            if (nativeContainer != null) {
                ComponentPeer peer = nativeContainer.getPeer();
                if (peer != null) {
                    peer.handleEvent(currentEventAndMostRecentTime);
                    currentEventAndMostRecentTime.consume();
                }
            }
            return true;
        }
        else {
            List<KeyEventDispatcher> keyEventDispatchers = this.getKeyEventDispatchers();
            if (keyEventDispatchers != null) {
                Iterator<Object> iterator = keyEventDispatchers.iterator();
                while (iterator.hasNext()) {
                    if (!((KeyEventDispatcher)iterator.next()).dispatchKeyEvent(currentEventAndMostRecentTime)) {
                        continue;
                    }
                    return true;
                }
            }
            return this.dispatchKeyEvent(currentEventAndMostRecentTime);
        }
    }
    
    private void consumeNextKeyTyped(KeyEvent keyEvent) {
        this.consumeNextKeyTyped = true;
    }
    
    private void consumeTraversalKey(KeyEvent keyEvent) {
        keyEvent.consume();
        this.consumeNextKeyTyped = (keyEvent.getID() == 401 && !keyEvent.isActionKey());
    }
    
    private boolean consumeProcessedKeyEvent(KeyEvent keyEvent) {
        if (keyEvent.getID() == 400 && this.consumeNextKeyTyped) {
            keyEvent.consume();
            this.consumeNextKeyTyped = false;
            return true;
        }
        return false;
    }
    
    @Override
    public void processKeyEvent(Component component, KeyEvent keyEvent) {
        if (this.consumeProcessedKeyEvent(keyEvent)) {
            return;
        }
        if (keyEvent.getID() == 400) {
            return;
        }
        if (component.getFocusTraversalKeysEnabled() && !keyEvent.isConsumed()) {
            AWTKeyStroke awtKeyStrokeForEvent = AWTKeyStroke.getAWTKeyStrokeForEvent(keyEvent);
            AWTKeyStroke awtKeyStroke = AWTKeyStroke.getAWTKeyStroke(awtKeyStrokeForEvent.getKeyCode(), awtKeyStrokeForEvent.getModifiers(), !awtKeyStrokeForEvent.isOnKeyRelease());
            Set<AWTKeyStroke> focusTraversalKeys = component.getFocusTraversalKeys(0);
            int contains = focusTraversalKeys.contains(awtKeyStrokeForEvent);
            int contains2 = focusTraversalKeys.contains(awtKeyStroke);
            if (contains || contains2) {
                this.consumeTraversalKey(keyEvent);
                if (!(contains == 0)) {
                    this.focusNextComponent(component);
                }
                return;
            }
            else {
                if (keyEvent.getID() == 401) {
                    this.consumeNextKeyTyped = false;
                }
                Set<AWTKeyStroke> focusTraversalKeys2 = component.getFocusTraversalKeys(1);
                int contains3 = focusTraversalKeys2.contains(awtKeyStrokeForEvent);
                int contains4 = focusTraversalKeys2.contains(awtKeyStroke);
                if (contains3 || contains4) {
                    this.consumeTraversalKey(keyEvent);
                    if (!(contains3 == 0)) {
                        this.focusPreviousComponent(component);
                    }
                    return;
                }
                else {
                    Set<AWTKeyStroke> focusTraversalKeys3 = component.getFocusTraversalKeys(2);
                    int contains5 = focusTraversalKeys3.contains(awtKeyStrokeForEvent);
                    int contains6 = focusTraversalKeys3.contains(awtKeyStroke);
                    if (contains5 || contains6) {
                        this.consumeTraversalKey(keyEvent);
                        if (!(contains5 == 0)) {
                            this.upFocusCycle(component);
                        }
                        return;
                    }
                    else {
                        if (!(component instanceof Container) || !((Container)component).isFocusCycleRoot()) {
                            return;
                        }
                        Set<AWTKeyStroke> focusTraversalKeys4 = component.getFocusTraversalKeys(3);
                        int contains7 = focusTraversalKeys4.contains(awtKeyStrokeForEvent);
                        int contains8 = focusTraversalKeys4.contains(awtKeyStroke);
                        if (contains7 || contains8) {
                            this.consumeTraversalKey(keyEvent);
                            if (!(contains7 == 0)) {
                                this.downFocusCycle((Container)component);
                            }
                        }
                    }
                }
            }
        }
        return;
    }
    
    @Override
    protected synchronized void enqueueKeyEvents(long l, Component component) {
        if (component == null) {
            return;
        }
        if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
            DefaultKeyboardFocusManager.focusLog.finer("Enqueue at {0} for {1}", new Object[] { Long.valueOf(l), component });
        }
        int index = 0;
        int i = this.typeAheadMarkers.size();
        ListIterator<DefaultKeyboardFocusManager.TypeAheadMarker> listIterator = this.typeAheadMarkers.listIterator(i);
        while (i > 0) {
            if (((DefaultKeyboardFocusManager.TypeAheadMarker)listIterator.previous()).after <= l) {
                index = i;
                break;
            }
            --i;
        }
        this.typeAheadMarkers.add(index, new DefaultKeyboardFocusManager.TypeAheadMarker(l, component));
        return;
    }
    
    @Override
    protected synchronized void dequeueKeyEvents(long l, Component component) {
        if (component == null) {
            return;
        }
        if (DefaultKeyboardFocusManager.focusLog.isLoggable(PlatformLogger.Level.FINER)) {
            DefaultKeyboardFocusManager.focusLog.finer("Dequeue at {0} for {1}", new Object[] { Long.valueOf(l), component });
        }
        ListIterator<DefaultKeyboardFocusManager.TypeAheadMarker> listIterator = this.typeAheadMarkers.listIterator((l >= 0L) ? this.typeAheadMarkers.size() : 0);
        if (l < 0L) {
            while (listIterator.hasNext()) {
                if (((DefaultKeyboardFocusManager.TypeAheadMarker)listIterator.next()).untilFocused != component) {
                    continue;
                }
                listIterator.remove();
                return;
            }
        }
        else {
            while (listIterator.hasPrevious()) {
                DefaultKeyboardFocusManager.TypeAheadMarker typeAheadMarker = (DefaultKeyboardFocusManager.TypeAheadMarker)listIterator.previous();
                if (typeAheadMarker.untilFocused != component || typeAheadMarker.after != l) {
                    continue;
                }
                listIterator.remove();
                return;
            }
        }
        return;
    }
    
    @Override
    protected synchronized void discardKeyEvents(Component component) {
        if (component == null) {
            return;
        }
        long after = -1L;
        Iterator<Object> iterator = this.typeAheadMarkers.iterator();
        while (iterator.hasNext()) {
            DefaultKeyboardFocusManager.TypeAheadMarker typeAheadMarker = (DefaultKeyboardFocusManager.TypeAheadMarker)iterator.next();
            Component component2;
            int n;
            for (component2 = typeAheadMarker.untilFocused, n = (component2 == component); n == 0 && component2 != null && !(component2 instanceof Window); component2 = component2.getParent(), n = (component2 == component)) {}
            if (n != 0) {
                if (after < 0L) {
                    after = typeAheadMarker.after;
                }
                iterator.remove();
            }
            else if (after >= 0L) {
                this.purgeStampedEvents(after, typeAheadMarker.after);
                after = -1L;
            }
            else {
                continue;
            }
        }
        this.purgeStampedEvents(after, -1L);
        return;
    }
    
    private void purgeStampedEvents(long n, long n2) {
        if (n < 0L) {
            return;
        }
        Iterator<Object> iterator = this.enqueuedKeyEvents.iterator();
        while (iterator.hasNext()) {
            long when = ((KeyEvent)iterator.next()).getWhen();
            if (n < when && (n2 < 0L || when <= n2)) {
                iterator.remove();
            }
            if (n2 >= 0L && when > n2) {
                break;
            }
            continue;
        }
        return;
    }
    
    @Override
    public void focusPreviousComponent(Component component) {
        if (component != null) {
            component.transferFocusBackward();
        }
    }
    
    @Override
    public void focusNextComponent(Component component) {
        if (component != null) {
            component.transferFocus();
        }
    }
    
    @Override
    public void upFocusCycle(Component component) {
        if (component != null) {
            component.transferFocusUpCycle();
        }
    }
    
    @Override
    public void downFocusCycle(Container container) {
        if (container != null && container.isFocusCycleRoot()) {
            container.transferFocusDownCycle();
        }
    }
    
    static void access$000(DefaultKeyboardFocusManager defaultKeyboardFocusManager, KeyEvent keyEvent) {
        defaultKeyboardFocusManager.consumeNextKeyTyped(keyEvent);
    }
    
    static {
        focusLog = PlatformLogger.getLogger("java.awt.focus.DefaultKeyboardFocusManager");
        NULL_WINDOW_WR = new WeakReference(null);
        NULL_COMPONENT_WR = new WeakReference(null);
        AWTAccessor.setDefaultKeyboardFocusManagerAccessor(new AWTAccessor.DefaultKeyboardFocusManagerAccessor() {
            @Override
            public void consumeNextKeyTyped(DefaultKeyboardFocusManager defaultKeyboardFocusManager, KeyEvent keyEvent) {
                DefaultKeyboardFocusManager.access$000(defaultKeyboardFocusManager, keyEvent);
            }
        });
    }
    
    private static class TypeAheadMarker
    {
        long after;
        Component untilFocused;
        
        TypeAheadMarker(long after, Component untilFocused) {
            this.after = after;
            this.untilFocused = untilFocused;
        }
        
        @Override
        public String toString() {
            return ">>> Marker after " + this.after + " on " + this.untilFocused;
        }
    }
    
    private static class DefaultKeyboardFocusManagerSentEvent extends SentEvent
    {
        private static final long serialVersionUID = -2924743257508701758L;
        
        public DefaultKeyboardFocusManagerSentEvent(AWTEvent awtEvent, AppContext appContext) {
            super(awtEvent, appContext);
        }
        
        @Override
        public final void dispatch() {
            KeyboardFocusManager currentKeyboardFocusManager = KeyboardFocusManager.getCurrentKeyboardFocusManager();
            DefaultKeyboardFocusManager defaultKeyboardFocusManager = (currentKeyboardFocusManager instanceof DefaultKeyboardFocusManager) ? ((DefaultKeyboardFocusManager)currentKeyboardFocusManager) : null;
            if (defaultKeyboardFocusManager != null) {
                synchronized (defaultKeyboardFocusManager) {
                    defaultKeyboardFocusManager.inSendMessage++;
                }
            }
            super.dispatch();
            if (defaultKeyboardFocusManager != null) {
                synchronized (defaultKeyboardFocusManager) {
                    defaultKeyboardFocusManager.inSendMessage--;
                }
            }
        }
    }
}
